from .. import NiaInterface
from ..crossovers import SBX, RandomSBX
from ..mutations import Uniform
import numpy as np
from random import random

class GeneticAlgorithm(NiaInterface):
    @NiaInterface.initializer
    def __init__(self, 
                cost_function,
                lower_bond,
                upper_bond,
                iteration_function = None,
                num_population = 100,
                num_parents = 20,
                max_iteration = 100,
                num_variable = 1,
                quit_criterion = 0.001,
                crossover = RandomSBX(2),
                mutation = Uniform(0.05)
                ):
        self.lower_bond = np.array(lower_bond)
        self.upper_bond = np.array(upper_bond)

    def generate_population(self, volume):
        return np.random.rand(volume, self.num_variable) * (self.upper_bond - self.lower_bond) + self.lower_bond

    def get_fitness(self, population):
        fitness = np.zeros(shape=(len(population),1))
        for i ,pop in enumerate(population):
            fitness[i] = self.cost_function(pop)
        return fitness

    def sort_by_fitness(self, population, fitness):
        new_population = np.append(population, fitness, axis=1)
        new_population = new_population[new_population[:, -1].argsort()]
        return new_population[:,:-1], new_population[:,-1:]

    def run(self):
        self.population = self.generate_population(self.num_population)
        self.fitness = self.get_fitness(self.population)
        self.population, self.fitness = self.sort_by_fitness(self.population, self.fitness)
        for self.iteration in range(self.max_iteration):
            children = self.crossover.generate(self.population[:self.num_parents], self.num_parents)
            children = self.mutation.mutate(children, self.generate_population(self.num_parents))
            children_fittness = self.get_fitness(children)            
            self.population , self.fitness = self.sort_by_fitness(
                np.concatenate((self.population, children), axis=0),
                np.concatenate((self.fitness, children_fittness), axis=0)
            )
            self.population = self.population[:self.num_population]
            self.fitness = self.fitness[:self.num_population]
            self.best = (self.population[0], self.fitness[0])
            if self.iteration_function:
                self.iteration_function(self)
            if self.fitness[0][0] < self.quit_criterion:
                break
        return self.best
        


from .. import NiaInterface
import numpy as np
from random import random
from copy import copy

class ParticleSwarmOptimization(NiaInterface):
    @NiaInterface.initializer
    def __init__(self, 
                cost_function,          #fitness function
                iteration_function,     #will be called in each iteration
                lower_bond,             #lower band of parameters
                upper_bond,             #upper band of parameters
                num_particles = 10,        #number of particles
                max_iterations = 100,   #max number of iterations
                num_variable = 1,       #number of parameters to be optimized
                c1 = 0.1,               #cognitive coefficent ,c1>=0
                c2 = 0.1,               #social coefficent    ,c2<=4
                quit_criteria = 0.0001, #acceptable fitness value
                ):
        self.lower_bond = np.array(lower_bond)
        self.upper_bond = np.array(upper_bond)
        self.fitness = np.zeros(shape=(num_particles,1))
        
    class Particle:
        pass

    def generate_particles(self):
        self.particles = []
        for i in range(self.num_particles):
            particle = self.Particle()
            particle.position = np.array([random() for i in range(self.num_variable)]) * (self.upper_bond - self.lower_bond) + self.lower_bond
            particle.fitness  = self.calculate_fitness(particle)
            particle.velocity = 0.0
            particle.personal_best     = particle.fitness
            self.particles.append(particle)

    def calculate_fitness(self,particle):
        return self.cost_function(particle.position)
        
    def log(self):
        print(self.population)
        print(self.fitness)

    def run(self):
        self.generate_particles()
        self.best_particle = copy(self.particles[0])
        for i in range(self.max_iterations) :
            for particle in self.particles:
                fitness = self.calculate_fitness(particle)
                if fitness < particle.fitness:
                    particle.personal_best = particle.position
                particle.fitness = fitness

                if fitness < self.best_particle.fitness:
                    self.best_particle = copy(particle)
                
                particle.velocity = particle.velocity + self.c1 * random() * (particle.personal_best - particle.position) \
                                 + self.c2 * random() * (self.best_particle.position - particle.position)
                particle.position = particle.position + particle.velocity

            self.iteration_function(self)
            if fitness < self.quit_criteria:
                self.message = 'quit criteria reached best answer is: ' + str(self.best_particle.position) + ' and best fitness is: ' + str(self.best_particle.fitness) + ' iteration : ' + str(i);
                return self
        self.message = 'max itteration reached last answer is : ' + str(self.best_particle.position) + ' and best fitness is: ' + str(self.best_particle.fitness);
        return self
        









from .mutation import Mutation
import numpy as np

class Uniform(Mutation):
    @Mutation.initializer
    def __init__(self, rate=0.01):
        pass

    def mutate(self, population, random_population):
        random_rates = np.random.rand(len(population), len(population[0])) # check this
        random_population_zeros = np.where(random_rates >= self.rate, 0, 1)
        population_zeros = np.where(random_rates < self.rate, 0, 1)
        random_population = random_population * random_population_zeros
        population = population * population_zeros
        return population + random_population


from .crossover import CrossOver
import random
import numpy as np

class SBX(CrossOver):
    @CrossOver.initializer
    def __init__(self, eta=2):
        pass

    def generate(self, population, num_children):
        children = []
        for i in range(int(len(population)/2)):
            father1, father2 = population[i], population[i+1]
            rand = random.random()
            rand = 1 - rand if rand >= 0.5 else rand
            beta = (2 * rand) ** (1 / (1 + self.eta))
            son1 = 0.5 * (1 + beta) * father1 + (1 - beta) * father2
            son2 = 0.5 * (1 - beta) * father1 + (1 + beta) * father2
            children += [son1, son2]
        return np.array(children)